#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: Seven Yu
# @Date: May 8, 2022
# @Function: Denoising 4D fMRI BOLD file
# Copyright:
#    Made by Seven Yu with ♥ 
#    yusaiwen@mail.bnu.edu.cn,
#    State Key Laboratory of Cognitive Neuroscience and Learning,
#    Beijing Normal University,
#    No.19 Xinjiekouwai Street, Haidian District,
#    Beijing, China, 100875

VERSION = '0.1.0'
NAME = 'fMRI Denoiser'
ACCOUNT = 'Denoise fMRI NIFTI file following fMRIPREP'
COMMAND_NAME = 'denoiser'
MEME = "💦"

from rich.console import Console

console = Console()

import .RC
# load basic packages
from rich.color import Color
from rich.text import Text
from termcolor import colored

#import rich_click as click
import click

from os.path import exists as ope
from os.path import abspath as opa
from os.path import join as opj
import os
import sys

import logging
from rich.logging import RichHandler


from bids import BIDSLayout
from nipype import DataSink, SelectFiles, IdentityInterface, Node, Workflow
from nipype.interfaces import afni

import warnings
with warnings.catch_warnings():
    # silent nilearn.image
    warnings.filterwarnings("ignore", category=RuntimeWarning)
    from nilearn import image as nimg
from nilearn.interfaces.fmriprep import load_confounds as load_confounds
from nilearn.image import clean_img as denoiser

import numpy as np



class NoSuchPath(Exception):
    '''
    An self-made exception: No such path
    '''
    pass

def yaml2dic(yml_file):
    import yaml 

    with open(yml_file,'r', encoding='UTF-8') as f:
        return yaml.load(f, Loader=yaml.Loader)


def red_error(account):
    return '[bold red]' + account + '[/]'

def get_task(layout):
    return layout.get_task()
def get_run(layout):
    return layout.get_run()
def get_subjects(layout):
    from natsort import natsorted as ns
    return ns(layout.get_subjects())
def get_TR(layout):
    return layout.get_tr()

def get_bids_meta(layout, info):
    switcher = {
        'sub_list': get_subjects(layout),
        'task': get_task(layout),
        'run': get_run(layout),
        'TR': get_TR(layout)
    }
    return switcher.get(info)
    

def get_fmri(sub, layout, task, run, log):        
    fmri_file = layout.get(
        scope='derivatives',
        subject=sub,
        task=task,
        run=run,
        space='MNI152NLin2009cAsym',
        return_type='file',
        extension='nii.gz',
        suffix='bold'
    )
    if len(fmri_file) != 1:
        try:
            raise ValueError
        except:
            error_account = 'Multiple fMRI file or none fMRI file was found (／‵Д′)／~ ╧╧'
            log.exception(red_error(error_account), extra={"markup": True})
            sys.exit(1)
    return fmri_file[0]

def get_mask(sub, layout, task, run):
    return layout.get(
        scope='derivatives',
        subject=sub,
        task=task,
        run=run,
        space='MNI152NLin2009cAsym',
        return_type='file',
        extension='nii.gz',
        suffix='mask'
    )[0]

def get_confounds(fmri_file, confounds):
    with warnings.catch_warnings():
    # silent nilearn.image
        warnings.filterwarnings("ignore", category=UserWarning)
        return load_confounds(
            # this strategy contains 36 parameters
            # to be updated!
            img_files=fmri_file,
            strategy=confounds.get('strategy'),
            motion=confounds.get('motion'),
            wm_csf=confounds.get('wm_csf'),
            global_signal=confounds.get('global_signal')
        )

def denoise_run(sub, layout, task, run, confounds, TR, pass_band, log):
    # get two files
    # * 4D fMRI NiFTI file
    # * mask
    # * confounds
    fmri_file = get_fmri(sub, layout, task, run, log)
    mask_file = get_mask(sub, layout, task, run)
    confounds = get_confounds(fmri_file, confounds)

    log.info(f"fMRI file: {fmri_file} \nfMRI mask: {mask_file}")

    # drop the dummy scans
    raw_func_img = nimg.load_img(fmri_file)
    func_img = raw_func_img.slicer[:,:,:,4:] # 4th dim is time, d with the first four volumes
    mask_img = nimg.load_img(mask_file)

    # denoising!
    denoised_img = denoiser(
        imgs=func_img,
        confounds=confounds[0].loc[4:], # confounds should be trimed as well
        low_pass=pass_band[0],
        high_pass=pass_band[1],
        t_r=TR,
        mask_img=mask_img
    )
    denoise_tmp_dir = opa('/tmp/denoising')
    denoise_tmp_file = opj(denoise_tmp_dir, 'sub-'+str(sub)+'_task-'+task+'_run-'+str(run)+'_denoised_img.nii.gz')
    denoised_img.to_filename(denoise_tmp_file)

RC.VERSION, RC.NAME, RC.ACCOUNT, RC.COMMAND_NAME, RC.MEME = [VERSION, NAME, ACCOUNT, COMMAND_NAME, MEME]
@click.command(cls=RC.RichCommand)
@click.option('--config', '-c',
              required=True,
              type=str,
              help='General YAML configuration file.'
             )
@click.option('--denoise', '-d',
              required=True,
              type=str,
              help='Denoise YAML configuration file.'
             )
@click.option('--log_level', '-l',
              default='INFO',
              show_default=True,
              metavar='CRITICAL, ERROR, WARNING, INFO, DEBUG, NOTSET',
              type=click.Choice(['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG', 'NOTSET']),
              help="Logging level."
             )
@click.option("--version", "-v",
              is_flag=True,
              help="Print version and exit."
             )
################ main function ############
def main(config,
         denoise,
         log_level: str = 'INFO',
         version: bool = False
         ):

    console = Console()

    if version:
        """
            print the version and leave
        """
        console.print(f"{VERSION}\n")
        sys.exit(1)

    # configure log package
    logging.basicConfig(
        level=eval(f"logging.{log_level}"),
        format="%(message)s",
        datefmt="[%X]",
        handlers=[RichHandler(rich_tracebacks=True)]
    )
    log = logging.getLogger("rich")

    # read general configuration file
    ge_config = opa(config)

    if not ope(ge_config):
        # if json file exists ?
        try:
            raise NoSuchPath
        except:
            error_account = 'Can\'t find your general configuration file (／‵Д′)／~ ╧╧'
            log.exception(red_error(error_account), extra={"markup": True})
            sys.exit(1)

    bids_dir, layout_dir = list(yaml2dic(ge_config).values())

    # read denoise configuration file
    de_config = opa(denoise)

    if not ope(de_config):
        # if json file exists ?
        try:
            raise NoSuchPath
        except:
            error_account = 'Can\'t find your denoising configuration file (／‵Д′)／~ ╧╧'
            log.exception(red_error(error_account), extra={"markup": True})
            sys.exit(1)

    confounder, pass_band, dummy = list(yaml2dic(de_config).values())

    # load pyBIDS
    layout_dir = opj(bids_dir, layout_dir)
    if not ope(layout_dir):
        try:
            raise NoSuchPath
        except:
            error_account = 'Can\'t find your layout path (／‵Д′)／~ ╧╧'
            log.exception(red_error(error_account), extra={"markup": True})
            sys.exit(1)
    else:
        console.log(f"Layout directory {layout_dir} found!")

    layout = BIDSLayout.load(layout_dir)

    # Get BIDS meta info
    info_list = ['sub_list', 'task', 'run', 'TR']
    info_dic = {}

    with console.status("[bold green]Get BIDS meta info...", spinner='moon') as status:
        while info_list:
            info = info_list.pop(0)
            info_dic[info] = get_bids_meta(layout, info)
            log.info(f"Fetched {info}: {info_dic.get(info)}!")
    
    # if just one run(run list is empty), run should be [None]
    if not len(info_dic.get('run')):
        info_dic['run'] = [None]
    sys.exit(1)
    with console.status("[bold green]Denoising...", spinner='moon') as status:
        #for sub in [1]: 
        for sub in info_dic.get('sub_list'):
            for task in info_dic.get('task'):
                for run in info_dic.get('run'):
                    denoise_run(sub=sub,
                                layout=layout,
                                task=task,
                                run=run,
                                TR = info_dic.get('TR'),
                                pass_band = denoise.get('pass_band'),
                                confounds=denoise.get('confounder'),
                                log=log)
            console.log(f"sub {sub} complete")


if __name__ == '__main__':
    main()
    console = Console()
    cong_account = 'Great! You\'ve done everything without error!  (ﾉ>ω<)ﾉ'
    console.log(cong_account, style='bold yellow')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fmricat']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'click>=8.1.3,<9.0.0',
 'nilearn>=0.9.1,<0.10.0',
 'nipype>=1.6.9,<2.0.0',
 'numpy>=1.22.3,<2.0.0',
 'pybids>=0.15.1,<0.16.0',
 'rich>=12.4.0,<13.0.0',
 'scipy==1.8.0',
 'sklearn>=0.0,<0.1',
 'termcolor>=1.1.0,<2.0.0']

entry_points = \
{'console_scripts': ['denoiser = fmricat.denoiser:main()']}

setup_kwargs = {
    'name': 'fmricat',
    'version': '0.0.1',
    'description': 'An automated fMRI analysis toolbox',
    'long_description': None,
    'author': 'seven',
    'author_email': 'yusaiwen@mail.bnu.edu.cn',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '==3.10.2',
}


setup(**setup_kwargs)



import click

from rich.highlighter import RegexHighlighter
from rich.panel import Panel
from rich.table import Table
from rich.theme import Theme
from rich.color import Color
from rich.text import Text
from rich.console import Console

from typing import TYPE_CHECKING, List, NoReturn, Optional, Tuple

def blend_text(
    message: str, color1: Tuple[int, int, int], color2: Tuple[int, int, int]
) -> Text:
    """Blend text from one color to another."""
    text = Text(message)
    r1, g1, b1 = color1
    r2, g2, b2 = color2
    dr = r2 - r1
    dg = g2 - g1
    db = b2 - b1
    size = len(text)
    for index in range(size):
        blend = index / size
        color = f"#{int(r1 + dr * blend):2X}{int(g1 + dg * blend):2X}{int(b1 + db * blend):2X}"
        text.stylize(color, index, index + 1)
    return text

class RichCommand(click.Command):
    """Override Clicks help with a Richer version."""

    def format_help(self, ctx, formatter):
        class OptionHighlighter(RegexHighlighter):
            highlights = [
                r"(?P<switch>\-\w)",
                r"(?P<option>\-\-[\w\-]+)",
            ]

        highlighter = OptionHighlighter()

        console = Console(
            theme=Theme(
                {
                    "option": "bold cyan",
                    "switch": "bold green",
                }
            ),
            highlighter=highlighter,
        )

        console.print(
            f"[b]{NAME}[/b] [magenta]v{VERSION}[/] {MEME}\n\n[dim]{ACCOUNT}.\n",
            justify="center",
        )

        console.print(
            f"Usage: [b]{COMMAND_NAME}[/b] [b][OPTIONS][/] \n"
        )

        options_table = Table(highlight=True, box=None, show_header=False)
        
        for param in self.get_params(ctx)[0:]:
            if len(param.opts) == 2:
                opt1 = highlighter(param.opts[1])
                opt2 = highlighter(param.opts[0])
            else:
                opt2 = highlighter(param.opts[0])
                opt1 = Text("")

            if type(param.type) == click.types.Choice:
                opt2 += Text(" "+"|".join(param.type.to_info_dict().get('choices')), style="bold yellow")

            options = Text(" ".join(reversed(param.opts)))
            help_record = param.get_help_record(ctx)
            if help_record is None:
                help = ""
            else:
                help = Text.from_markup(param.get_help_record(ctx)[-1], emoji=False)
                if param.show_default:
                    help += Text(f"Default: {param.default}", style="dim")
                

            if param.metavar:
                options += f" {param.metavar}"

            
            #     options += f" {param.default}"
                    

            options_table.add_row(opt1, opt2, highlighter(help))

        console.print(
            Panel(
                options_table, border_style="dim", title="Options", title_align="left"
            )
        )

        console.print(
            blend_text(
                "🇨🇳  Made by Seven Yu with ♥\n🏠 yusaisai.top\n💌 yusaiwen@mail.bnu.edu.cn",
                Color.parse("#b169dd").triplet,
                Color.parse("#542c91").triplet,
            ),
            justify="left",
            style="bold",
        )